from django.contrib.auth import get_user_model

User = get_user_model()


class GARBackend:
    """
    CAS authentication with UAI (Unité Administrative Immatriculée) number
    """

    @staticmethod
    def authenticate(request, uai_numbers):
        user = User.objects.filter(
            garinstitution__uai__in=uai_numbers, is_active=True
        ).last()
        return user

    @staticmethod
    def get_user(user_id):
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None



import logging
import pytest

from django_gar.views import LogoutView

pytestmark = pytest.mark.django_db


class TestLogoutView:
    def test_with_unknown_cas_ticket(self, mocker, logout_body, request_builder):
        # GIVEN
        mock_logger = mocker.patch.object(logging.getLogger("django_gar.views"), "info")
        request = request_builder.post(data=logout_body)

        # WHEN
        response = LogoutView.as_view()(request)

        # THEN
        assert response.status_code == 204
        assert mock_logger.call_count == 2
        logger_messages = [call[0] for call, _ in mock_logger.call_args_list]
        assert logger_messages[0] == logout_body
        assert (
            logger_messages[1]
            == "cannot log out from GAR session as it does not exist anymore"
        )

    def test_with_existing_gar_session(
        self, mocker, logout_body, request_builder, gar_session
    ):
        # GIVEN
        mock_logger = mocker.patch.object(logging.getLogger("django_gar.views"), "info")
        request = request_builder.post(data=logout_body)

        # WHEN
        response = LogoutView.as_view()(request)

        # THEN
        assert response.status_code == 204
        assert mock_logger.call_count == 2
        logger_messages = [call[0] for call, _ in mock_logger.call_args_list]
        assert logger_messages[0] == logout_body
        assert logger_messages[1] == f"request GAR logout {gar_session.session_key}"



import time
import requests
import logging
from bs4 import BeautifulSoup
from datetime import datetime
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db.models.signals import pre_save, post_delete
from django.dispatch import receiver

from defusedxml import ElementTree as ET

from ..models import GARInstitution
from ..gar import (
    delete_gar_subscription,
    get_gar_certificate,
    get_gar_headers,
    get_gar_request_url,
    get_gar_institution_list,
)

logger = logging.getLogger(__name__)

GAR_DISTRIBUTOR_ID = getattr(settings, "GAR_DISTRIBUTOR_ID", "")
GAR_RESOURCES_ID = getattr(settings, "GAR_RESOURCES_ID", "")
GAR_ORGANIZATION_NAME = getattr(settings, "GAR_ORGANIZATION_NAME", "")
GAR_SUBSCRIPTION_PREFIX = getattr(settings, "GAR_SUBSCRIPTION_PREFIX", "")


@receiver(post_delete, sender=GARInstitution, dispatch_uid="delete_subscription_in_gar")
def delete_subscription_in_gar(sender, instance, **kwargs):
    delete_gar_subscription(instance.subscription_id)


def _get_gar_start_date(instance, http_method):
    """Get the start date for GAR subscription"""
    if http_method == "POST":
        uai = instance.uai.upper().strip()
        data = """<?xml version="1.0" encoding="UTF-8"?>
        <filtres xmlns="http://www.atosworldline.com/wsabonnement/v1.0/">
              <filtre>
                    <filtreNom>idDistributeurCom</filtreNom>
                    <filtreValeur>{distributor_id}</filtreValeur>
              </filtre> 
              <filtre>
                    <filtreNom>uaiEtab</filtreNom>
                    <filtreValeur>{uai}</filtreValeur>
              </filtre> 
        </filtres>""".format(
            distributor_id=GAR_DISTRIBUTOR_ID, uai=uai
        )
        cert = get_gar_certificate()
        headers = get_gar_headers()
        response = requests.request(
            "GET",
            "https://abonnement.gar.education.fr/abonnements",
            data=data,
            cert=cert,
            headers=headers,
        )
        soup = BeautifulSoup(response.text, "lxml")
        subscriptions = soup.findAll("abonnement")
        for subscription in subscriptions:
            if subscription.find("idabonnement").text == instance.subscription_id:
                return subscription.find("debutvalidite").text

    return datetime.now().isoformat()


def _get_gar_data_to_send(instance, http_method=None):
    """Generate XML data for GAR subscription"""
    uai = instance.uai.upper().strip()
    xml = """<?xml version="1.0" encoding="UTF-8"?>
    <abonnement xmlns="http://www.atosworldline.com/wsabonnement/v1.0/">
       <idAbonnement>{subscription_id}</idAbonnement>
       <idDistributeurCom>{distributor_id}</idDistributeurCom>
       <idRessource>{resources_id}</idRessource>
       <typeIdRessource>ark</typeIdRessource>
       <libelleRessource>{organization_name}</libelleRessource>
       <debutValidite>{start_date}</debutValidite>
       <finValidite>{end_date}T23:59:59</finValidite>
       <uaiEtab>{uai}</uaiEtab>
       <categorieAffectation>transferable</categorieAffectation>
       <typeAffectation>INDIV</typeAffectation>
       <nbLicenceGlobale>ILLIMITE</nbLicenceGlobale>
       <publicCible>ELEVE</publicCible>
       <publicCible>ENSEIGNANT</publicCible>
       <publicCible>DOCUMENTALISTE</publicCible>
       <publicCible>AUTRE PERSONNEL</publicCible>
       <codeProjetRessource>{project_code}</codeProjetRessource>
    </abonnement>""".format(
        subscription_id=instance.subscription_id,
        distributor_id=GAR_DISTRIBUTOR_ID,
        resources_id=GAR_RESOURCES_ID,
        organization_name=GAR_ORGANIZATION_NAME,
        start_date=_get_gar_start_date(instance, http_method),
        end_date=instance.ends_at,
        uai=uai,
        project_code=instance.project_code,
    )

    if not instance.project_code:
        xml = xml.replace(
            f"<codeProjetRessource>{instance.project_code}</codeProjetRessource>", ""
        )

    if http_method == "POST":
        xml = xml.replace(f"<uaiEtab>{uai}</uaiEtab>", "")

    return xml


def _get_response_from_gar(instance, http_method):
    """Send request to GAR API"""
    url = get_gar_request_url(instance.subscription_id)
    cert = get_gar_certificate()
    headers = get_gar_headers()
    response = requests.request(
        http_method,
        url,
        data=_get_gar_data_to_send(instance, http_method=http_method),
        cert=cert,
        headers=headers,
    )

    if response.status_code == 409 and "existe deja" in response.text:
        response = _get_response_from_gar(instance, http_method="POST")

    return response


@receiver(pre_save, sender=GARInstitution, dispatch_uid="handle_gar_subscription")
def handle_gar_subscription(sender, instance, **kwargs):
    """Handle GAR subscription creation/update before saving the instance"""
    # Skip if we're only updating cache fields
    if kwargs.get('update_fields') == {'allocations_cache', 'allocations_cache_updated_at'} or \
       kwargs.get('update_fields') == {'subscription_cache', 'subscription_cache_updated_at'}:
        logger.info(f"GAR subscription caches for {instance.uai} updated")
        return

    if not instance.subscription_id:
        instance.subscription_id = f"{GAR_SUBSCRIPTION_PREFIX}{int(time.time())}"

    # Check if this is a new instance or an update
    if not instance.pk:
        response = _get_response_from_gar(instance, http_method="PUT")
        if response.status_code not in [201, 200]:
            logger.error(response.text)
    else:
        response = _get_response_from_gar(instance, http_method="POST")
        if response.status_code != 200:
            logger.error(response.text)

    logger.info(f"GAR subscription {instance.uai} updated with status code {response.status_code}")


@receiver(pre_save, sender=GARInstitution, dispatch_uid="get_id_ent")
def get_id_ent(sender, instance, **kwargs):
    if instance.pk:
        return

    institution_list = get_gar_institution_list()
    xml_data = institution_list.content
    root = ET.fromstring(xml_data)
    namespace = {"ns": "http://www.atosworldline.com/listEtablissement/v1.0/"}
    id_ent = None
    for etablissement in root.findall("ns:etablissement", namespace):
        uai = etablissement.find("ns:uai", namespace)
        if uai is not None and uai.text == instance.uai:
            # Found the correct UAI, now get the idENT
            id_ent_object = etablissement.find("ns:idENT", namespace)
            id_ent = id_ent_object.text if id_ent_object is not None else None
            continue
    if id_ent:
        instance.id_ent = id_ent
    else:
        logger.error(f"id ent not found for uai {instance.uai}")



import pytest
from django.utils import timezone

pytestmark = pytest.mark.django_db


class TestGARInstitutionSignals:
    def test_gar_api_not_called_for_cache_updates(self, user, mock_gar_request_response):
        # GIVEN
        institution = user.garinstitution
        
        # WHEN
        institution.allocations_cache = {"some": "data"}
        institution.allocations_cache_updated_at = timezone.now()
        institution.save(
            update_fields=["allocations_cache", "allocations_cache_updated_at"]
        )

        institution.subscription_cache = {"other": "data"}
        institution.subscription_cache_updated_at = timezone.now()
        institution.save(
            update_fields=["subscription_cache", "subscription_cache_updated_at"]
        )

        # THEN
        assert not mock_gar_request_response.called


import requests
import logging

from bs4 import BeautifulSoup
from urllib.parse import urlencode

from django.conf import settings

from .exceptions import DjangoGARException

logger = logging.getLogger(__name__)

GAR_SUBSCRIPTION_PREFIX = getattr(settings, "GAR_SUBSCRIPTION_PREFIX", "")
GAR_BASE_SUBSCRIPTION_URL = getattr(
    settings, "GAR_BASE_SUBSCRIPTION_URL", "https://abonnement.gar.education.fr/"
)
GAR_ALLOCATIONS_URL = getattr(
    settings,
    "GAR_ALLOCATIONS_URL",
    "https://decompte-affectations.gar.education.fr/decompteaffectations",
)
GAR_DISTRIBUTOR_ID = getattr(settings, "GAR_DISTRIBUTOR_ID", "")

GAR_CERTIFICATE_PATH = getattr(settings, "GAR_CERTIFICATE_PATH", "")
GAR_KEY_PATH = getattr(settings, "GAR_KEY_PATH", "")


def delete_gar_subscription(subscription_id):
    url = get_gar_request_url(subscription_id)
    cert = get_gar_certificate()
    headers = get_gar_headers()
    requests.delete(url, cert=cert, headers=headers, timeout=30)


def get_gar_request_url(subscription_id):
    base_url = GAR_BASE_SUBSCRIPTION_URL
    url = "{}{}".format(base_url, subscription_id)
    return url


def get_gar_certificate():
    cert = (GAR_CERTIFICATE_PATH, GAR_KEY_PATH)
    return cert


def get_gar_headers():
    headers = {"Content-Type": "application/xml"}
    return headers


def get_gar_subscription(uai, subscription_id):
    data = """<?xml version="1.0" encoding="UTF-8"?>
        <filtres xmlns="http://www.atosworldline.com/wsabonnement/v1.0/">
              <filtre>
                    <filtreNom>idDistributeurCom</filtreNom>
                    <filtreValeur>{distributor_id}</filtreValeur>
              </filtre> 
              <filtre>
                    <filtreNom>uaiEtab</filtreNom>
                    <filtreValeur>{uai}</filtreValeur>
              </filtre> 
        </filtres>""".format(
        distributor_id=GAR_DISTRIBUTOR_ID, uai=uai
    )
    response = requests.request(
        "GET",
        "{}{}".format(GAR_BASE_SUBSCRIPTION_URL, "abonnements"),
        data=data,
        cert=get_gar_certificate(),
        headers=get_gar_headers(),
    )

    if response.status_code != 200:
        raise DjangoGARException(
            status_code=response.status_code, message=response.text
        )

    soup = BeautifulSoup(response.text, "lxml")
    subscriptions = soup.findAll("abonnement")
    for subscription in subscriptions:
        if subscription.find("idabonnement").text == subscription_id:
            return subscription

    return None


def get_gar_subscription_end_date(uai, subscription_id):
    subscription = get_gar_subscription(uai, subscription_id)

    if subscription:
        return subscription.find("finvalidite").text

    return None


def get_gar_institution_list():
    return requests.request(
        "GET",
        f"{GAR_BASE_SUBSCRIPTION_URL}etablissements/etablissements.xml",
        cert=get_gar_certificate(),
        headers=get_gar_headers(),
    )


def get_allocations(subscription_id=None, project_code=None):
    if not subscription_id and not project_code:
        raise DjangoGARException(
            status_code=400,
            message="At least one of subscription_id or project_code is mandatory",
        )
    elif subscription_id and project_code:
        raise DjangoGARException(
            status_code=400,
            message="Cannot set subscription_id and project_code at the same time",
        )

    params = {"codeProjetRessource": project_code}
    if subscription_id:
        params = {"idAbonnement": subscription_id}

    return requests.request(
        "GET",
        f"{GAR_ALLOCATIONS_URL}?{urlencode(params)}",
        cert=get_gar_certificate(),
    )



# Generated by Django 2.2.15 on 2020-09-15 12:54

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_gar", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="garinstitution",
            name="subscription_id",
            field=models.CharField(
                default="temporary_id", max_length=255, verbose_name="id abonnement"
            ),
            preserve_default=False,
        ),
    ]



# Generated by Django 4.2.15 on 2024-08-30 09:00

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_gar", "0005_garinstitution_project_code"),
    ]

    operations = [
        migrations.AddField(
            model_name="garinstitution",
            name="id_ent",
            field=models.CharField(
                max_length=255, null=True, verbose_name="ID de l'ent"
            ),
        ),
    ]



import logging

from django.http import HttpResponseRedirect

from django.conf import settings
from django.contrib.auth import login, authenticate

from cas import CASClient
from defusedxml.ElementTree import fromstring, ParseError

from .models import GARSession

logger = logging.getLogger(__name__)

GAR_BASE_URL = getattr(settings, "GAR_BASE_URL", "")
GAR_ACTIVE_USER_REDIRECT = getattr(settings, "GAR_ACTIVE_USER_REDIRECT", "/")
GAR_INACTIVE_USER_REDIRECT = getattr(settings, "GAR_INACTIVE_USER_REDIRECT", "/")
GAR_QUERY_STRING_TRIGGER = getattr(settings, "GAR_QUERY_STRING_TRIGGER", "sso_id")


class GARMiddleware:
    """
    Middleware that allows CAS authentication with GAR
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        uai_number = request.GET.get(GAR_QUERY_STRING_TRIGGER, "")
        cas_ticket = request.GET.get("ticket", "")

        if cas_ticket and request.session.get("is_gar"):
            del request.session["is_gar"]
            uai_numbers = self.validate_ticket(request, cas_ticket)

            user = authenticate(request, uai_numbers=uai_numbers)
            if user:
                login(request, user, backend="django_gar.backends.GARBackend")

                # Session key might not exist at this point.
                # Calling save will force session to create a session_key
                request.session.save()
                GARSession.objects.update_or_create(
                    session_key=request.session.session_key,
                    defaults={"ticket": cas_ticket},
                )

                request.session["gar_user"] = True
                request.session["gar_uai"] = user.garinstitution.uai
                return HttpResponseRedirect(
                    request.GET.get("grain", GAR_ACTIVE_USER_REDIRECT)
                )
            else:
                return HttpResponseRedirect(GAR_INACTIVE_USER_REDIRECT)

        elif uai_number and uai_number.upper() == "GAR":
            url = self.get_cas_login_url(request)
            request.session["is_gar"] = True
            return HttpResponseRedirect(url)

        response = self.get_response(request)

        return response

    def get_cas_login_url(self, request):
        """Returns the CAS login url"""

        client = self.get_cas_client(request)

        return client.get_login_url()

    def validate_ticket(self, request, cas_ticket):
        """
        Validate the CAS ticket. Ticket lifetime is around 5 seconds.
        Returns the uai number if the user has access, None otherwise
        """

        client = self.get_cas_client(request)
        response = client.get_verification_response(cas_ticket)

        logger.info(response)

        try:
            tree = fromstring(response)
            ns = {"cas": "http://www.yale.edu/tp/cas"}
            auth_success_element = tree.find("cas:authenticationSuccess", ns)

            auth_success_element = auth_success_element.find("cas:attributes", ns)
            uai_element = "cas:UAI"

            uai_numbers = [
                uai.text.upper()
                for uai in auth_success_element.findall(uai_element, ns)
            ]

            request.session["gar_profile"] = auth_success_element.find("cas:PRO", ns).text

            request.session["gar_ido"] = auth_success_element.find("cas:IDO", ns).text

            return uai_numbers

        except (AttributeError, ParseError):
            return None

    @staticmethod
    def get_redirect_url(request):
        """Get redirect url for cas"""

        scheme = request.scheme
        host = request.get_host()
        url = "{}://{}/?{}={}".format(scheme, host, GAR_QUERY_STRING_TRIGGER, "gar")

        grain = request.GET.get("grain", "")
        if grain:
            url += f"&grain={grain}"

        return url

    def get_cas_client(self, request):
        """Create a CAS client"""
        service_url = self.get_redirect_url(request)

        client = CASClient(version=3, server_url=GAR_BASE_URL, service_url=service_url)

        return client



import pytest

from django.contrib.auth import get_user_model

from django_gar.backends import GARBackend

pytestmark = pytest.mark.django_db


class TestGARBackend(object):
    @staticmethod
    def test_user_has_access(user):
        """
        Test the case where user has access (is_active = True)
        """
        # WHEN
        backend = GARBackend()
        authenticated_user = backend.authenticate(
            request=None, uai_numbers=[user.garinstitution.uai]
        )

        # THEN
        assert authenticated_user is not None
        assert authenticated_user.email == user.email

    @staticmethod
    def test_user_doesnt_have_access(user):
        """
        Test the case where user doesn't have access (is_active = False)
        """
        # GIVEN
        user.is_active = False
        user.save()

        # WHEN
        backend = GARBackend()
        authenticated_user = backend.authenticate(
            request=None, uai_numbers=[user.garinstitution.uai]
        )

        # THEN
        assert authenticated_user is None

    @staticmethod
    def test_get_user_works(user):
        """
        Test the case where get_user find the corresponding user in the database
        """
        # WHEN
        backend = GARBackend()
        authenticated_user = backend.get_user(user_id=user.id)

        # THEN
        assert authenticated_user is not None

    @staticmethod
    def test_get_user_fails(user):
        """
        Test the case where get_user couldn't find the corresponding user in the database
        """
        # GIVEN
        UserModel = get_user_model()
        last_user = UserModel.objects.last()

        # WHEN
        backend = GARBackend()
        user = backend.get_user(user_id=last_user.id + 1)

        # THEN
        assert user is None



import pytest
import requests
from django.core.exceptions import ValidationError
from django_gar.forms import GARInstitutionForm
import datetime
import logging

pytestmark = pytest.mark.django_db


class TestGARInstitutionForm:
    def test_clean_uai_strips_and_uppercases(self):
        # GIVEN
        form = GARInstitutionForm(
            data={
                "uai": " abc ",
                "institution_name": "Test Institution",
                "ends_at": datetime.datetime.today(),
                "subscription_id": "test-123",
            }
        )

        # WHEN
        form.is_valid()  # This populates cleaned_data
        cleaned_uai = form.clean_uai()

        # THEN
        assert cleaned_uai == "ABC"

    def test_form_works_with_gar_when_creating_instance(
        self, form_data, mocker, response_from_gar
    ):
        # GIVEN
        mock_request = mocker.patch.object(
            requests,
            "request",
            return_value=response_from_gar(201, "dummy response message"),
        )

        # WHEN
        form = GARInstitutionForm(data=form_data().data)
        form.is_valid()
        form.save()

        # THEN
        assert mock_request.called_once()
        assert form.is_valid()

    def test_form_error_with_gar_when_creating_instance(
        self, form_data, mocker, response_from_gar, caplog
    ):
        # GIVEN
        mocker.patch.object(
            requests,
            "request",
            return_value=response_from_gar(400, "dummy error message"),
        )

        # WHEN
        form = GARInstitutionForm(data=form_data().data)
        form.is_valid()
        with caplog.at_level(logging.INFO):
            form.save()

        # THEN
        assert f"dummy error message" in caplog.records[0].message

    def test_form_works_with_gar_when_try_creating_instance_that_already_exists(
        self, form_data, mocker, response_from_gar
    ):
        # GIVEN
        mock_request = mocker.patch.object(
            requests,
            "request",
            side_effect=[
                response_from_gar(409, "Cette abonnement existe deja"),
                response_from_gar(200, "Hello"),
                response_from_gar(201, "OK"),
            ],
        )
        mocker.patch("django_gar.signals.handlers.handle_gar_subscription")

        # WHEN
        form = GARInstitutionForm(data=form_data().data)
        form.is_valid()
        form.save()

        # THEN
        assert mock_request.call_count == 3
        assert form.is_valid()

    def test_form_works_with_gar_when_updating_instance(
        self, form_data, mocker, response_from_gar, user
    ):
        # GIVEN
        institution = user.garinstitution
        data = form_data(garinstitution=institution).data
        mock_request = mocker.patch.object(
            requests,
            "request",
            return_value=response_from_gar(200, "dummy response message"),
        )
        mocker.patch("django_gar.signals.handlers.handle_gar_subscription")

        # WHEN
        form = GARInstitutionForm(instance=institution, data=data)
        form.is_valid()
        form.save()

        # THEN
        assert mock_request.called_once()
        assert form.is_valid()

    def test_form_error_with_gar_when_updating_instance(
        self, form_data, mocker, response_from_gar, user, caplog
    ):
        # GIVEN
        institution = user.garinstitution
        data = form_data(garinstitution=institution).data
        error_message = "dummy error message"
        mocker.patch.object(
            requests,
            "request",
            return_value=response_from_gar(400, error_message),
        )
        mocker.patch(
            "django_gar.signals.handlers.handle_gar_subscription",
            side_effect=ValidationError("GAR error"),
        )

        # WHEN
        form = GARInstitutionForm(instance=institution, data=data)
        form.is_valid()
        with caplog.at_level(logging.INFO):
            form.save()

        # THEN
        assert f"dummy error message" in caplog.records[0].message



import pytest

from django.db import IntegrityError

from .factories import InstitutionFactory

pytestmark = pytest.mark.django_db


class TestGARInstitutionModel(object):
    def test_garinstituion_has_unique_uai_number(self, user):
        uai = user.garinstitution.uai
        with pytest.raises(IntegrityError):
            InstitutionFactory(uai=uai, user=user)

    def test_garinstitution_str(self, user):
        institution = user.garinstitution
        assert str(institution) == "{} ({})".format(
            institution.institution_name, institution.uai
        )
